import pytest
import sys
sys.path.append('../scripts')
import cbdep.platform_introspection as plat

class TestPlatformIntrospection:

    @pytest.fixture(autouse=True)
    def cleanup(self):
        yield
        plat._processor = None
        plat._platforms = None

    def test__override_arch(self):
        plat.override_arch("arm64")
        assert plat.get_arches() == ["aarch64", "arm64", "aarch_64"]



"""
Dependency Management System
"""

import argparse
import cbdep
import importlib
import logging
import os
import os.path
import pathlib
import shutil
import sys
import yaml

from cbdep.cache import Cache
from cbdep.install import Installer
from cbdep.platform_introspection import get_arches, get_platforms, override_platforms, override_arch


# Set up logging and handler
logger = logging.getLogger('cbdep')
logger.setLevel(logging.DEBUG)
formatter = logging.Formatter('%(levelname)s: %(message)s')
handler = logging.StreamHandler()
handler.setFormatter(formatter)
handler.setLevel(logging.INFO)
logger.addHandler(handler)


class Cbdep:
    """

    """

    def __init__(self):
        cachedir = pathlib.Path.home() / ".cbdepcache"
        self.cache = Cache(str(cachedir))

    def do_cache(self, args):
        """
        Cache a URL
        """

        self.cache.get(args.url, args.recache)

        # Output the cache filename, if requested
        if args.report is not None:
            self.cache.report(args.url)

        # Save the cached file locally, if requested
        if args.output is not None:
            self.cache.save(args.url, args.output)

    @staticmethod
    def do_platform(self, args):
        """
        Display introspected platform information
        """

        logger.debug("Determining platform and arch...")
        print(get_platforms())
        print(get_arches())

    @staticmethod
    def loadconfig(args):
        """
        Returns the contents of the config file defining the available packages
        """

        yamlfile = args.config_file
        if yamlfile is not None:
            # Load text of file
            with open(yamlfile, 'r') as y:
                return y.read()

        return (importlib.resources.files(cbdep) / "cbdep.config").read_text()

    def do_install(self, args):
        """
        Install a package based on a descriptor YAML
        """

        installdir = args.dir
        if installdir is None:
            installdir = "install"
        installdir = str(pathlib.Path(installdir).resolve())

        installer = Installer.fromYaml(
            self.loadconfig(args),
            self.cache,
            get_platforms(),
            "x86" if args.x32 else get_arches()
        )
        installer.set_cache_only(args.cache_only)
        installer.set_recache(args.recache)
        if args.cache_local_file is not None:
            installer.set_from_local_file(args.cache_local_file)
            installer.set_cache_only(True)

        installer.install(
            args.package,
            args.version,
            args.base_url,
            installdir,
            args.cbdeps
        )

        if args.output is not None:
            logger.debug(f"Copying downloaded file to {args.output}")
            shutil.copy2(installer.get_installer_file(), args.output)

    def do_list(self, args):
        """
        List available packages
        """

        config = yaml.safe_load(self.loadconfig(args))
        pkgs = list(config['packages'].keys()) \
            + config['cbdeps']['packages']
        print(
            "Available packages (not all may be available on all platforms):"
        )
        for pkg in sorted(pkgs):
            print(f"  {pkg}")
        print()


def main():
    """
    """

    # PyInstaller binaries get LD_LIBRARY_PATH set for them, and that
    # can have unwanted side-effects for our own subprocesses. Remove
    # that here - it can still be set by a set_env: entry in cbdep.config
    # for an install directive.
    # This needs to be done very early - even the call to get_platforms()
    # below indirectly shells out to lsb_release.
    os.environ.pop("LD_LIBRARY_PATH", None)

    parser = argparse.ArgumentParser(
        description='Dependency Management System'
    )
    parser.add_argument(
        "--debug", action="store_true",
        help="Enable debugging output"
    )
    parser.add_argument(
        "-p", "--platform", type=str,
        default=None,
        help="Override detected platform (may be comma-separated list)"
    )
    parser.add_argument(
        "-a", "--arch", "--processor", type=str,
        default=None,
        help="Override detected architecture"
    )
    parser.add_argument(
        "-V", "--version", action="version",
        help="Display cbdep version information",
        version=f"cbdep version {importlib.metadata.version(__package__)}"
    )

    subparsers = parser.add_subparsers()

    cache_parser = subparsers.add_parser(
        "cache", help="Add downloaded URL to local cache"
    )
    cache_parser.add_argument("url", type=str, help="URL to cache")
    cache_parser.add_argument(
        "-r", "--report", action="store_true",
        help="Report the filename in the cache"
    )
    cache_parser.add_argument(
        "--recache", action="store_true",
        help="Re-download URL, replacing files in cache"
    )
    cache_parser.add_argument(
        "-o", "--output", type=str,
        help="Output cached file to a local file"
    )
    cache_parser.set_defaults(func=Cbdep.do_cache)

    install_parser = subparsers.add_parser(
        "install", help="Install a package"
    )
    install_parser.add_argument(
        "package", type=str, help="Package to install"
    )
    install_parser.add_argument(
        "version", type=str, help="Version to install"
    )
    install_parser.add_argument(
        "-3", "--x32", action="store_true",
        help="Download 32-bit package (default false; only works on "
             "a few packages)"
    )
    install_parser.add_argument(
        "-c", "--config-file", type=str,
        help="YAML file descriptor"
    )
    install_parser.add_argument(
        "-d", "--dir", type=str,
        help="Directory to unpack into (not applicable for all packages)"
    )
    install_parser.add_argument(
        "-b", "--base-url", type=str,
        help="Alternate base URL for downloading dep (only applicable to a few packages)"
    )
    install_parser.add_argument(
        "-C", "--cbdeps", action="store_true",
        help="Force interpreting 'package' as a cbdeps package"
    )
    install_parser.add_argument(
        "-n", "--cache-only", action='store_true',
        help="Only download any installer files, do not install"
    )
    install_parser.add_argument(
        "-r", "--report", action="store_true",
        help="Report the filename in the cache (only last-downloaded file"
            "in case of multiple downloads)"
    )
    install_parser.add_argument(
        "-o", "--output", type=str,
        help="Output cached file to a local file (only last-downloaded file"
            "in case of multiple downloads)"
    )
    install_parser.add_argument(
        "--recache", action="store_true",
        help="Re-download any installer files to cache, replacing files in cache"
    )
    install_parser.add_argument(
        "--cache-local-file", type=str,
        help="Populate cache with local file rather than downloading. Implies --cache-only."
    )
    install_parser.set_defaults(func=Cbdep.do_install)

    platform_parser = subparsers.add_parser(
        "platform", help="Dump introspected platform information"
    )
    platform_parser.set_defaults(func=Cbdep.do_platform)

    list_parser = subparsers.add_parser(
        "list", help="List available cbdep packages"
    )
    list_parser.add_argument(
        "-c", "--config-file", type=str, help="YAML file descriptor"
    )
    list_parser.set_defaults(func=Cbdep.do_list)

    args = parser.parse_args()

    # Set logging to debug level on stream handler if --debug was set
    if args.debug:
        handler.setLevel(logging.DEBUG)

    # Override architecture if specified
    if args.arch is not None:
        logger.debug(f"Overriding architecture to {args.arch}")
        override_arch(args.arch)

    # Override platform if specified
    if args.platform is not None:
        logger.debug(f"Overriding platform to {args.platform}")
        override_platforms(args.platform.split(','))

    # Check that a command was specified
    if "func" not in args:
        parser.print_help()
        sys.exit(1)

    cbdep = Cbdep()
    args.func(cbdep, args)


if __name__ == '__main__':
    main()



def __getattr__(name: str) -> str:
    if name != "__version__":
        msg = f"module {__name__} has no attribute {name}"
        raise AttributeError(msg)

    from importlib.metadata import version

    return version("cbdep")



"""
Install
"""

import logging
import os
import pathlib
import re
import shutil
import stat
import string
import sys
import tempfile
import yaml

from packaging.specifiers import SpecifierSet
from subprocess import run, CalledProcessError

import cbdep.zipfile_with_permissions as zipfile_with_permissions
from cbdep.platform_introspection import get_default_arches

logger = logging.getLogger("cbdep")
zipfile_with_permissions.register()


class Installer:
    """
    Manages caching installation files and unpacking them based on an
    installation template yaml file
    """

    def __init__(self, config, cache, platforms, arches):
        """
        Base constructor for class. "config" is a YAML object,
        "cache" is the cache directory, "platforms" is a list of
        platform names; "arches" is a list of architecture names.
        """

        # These fields are immutable and define the installation environment
        self.descriptor = config
        self.cache = cache
        self.platforms = platforms
        if not isinstance(self.platforms, list):
            self.platforms = [self.platforms]
        self.arches = arches
        if not isinstance(self.arches, list):
            self.arches = [self.arches]

        # Things that will be substituted in all templates
        self.symbols = dict()
        self.symbols["HOME"] = str(pathlib.Path.home())

        # Populated by install()
        self.package = None
        self.version = None
        self.safe_version = None
        self.base_url = None
        self.installdir = None

        # Default, can be overridden by self.set_cache_only()
        self.cache_only = False

        # Default, can be overridden by self.set_recache()
        self.recache = False

        # Default, can be overridden by self.set_from_local_file()
        self.from_local_file = None

        # Populated by do_url() to be the final single downloaded installer
        self.installer_file = None

        # Create a temp directory that action blocks can use
        self.temp_dir = tempfile.mkdtemp()
        self.symbols["TEMP_DIR"] = self.temp_dir

    def __del__(self):
        """
        Clean up
        """
        shutil.rmtree(self.temp_dir)

    @classmethod
    def fromYaml(cls, yamltext, cache, platforms, arches):
        """
        Constructor from a YAML configuration file
        """

        config = yaml.safe_load(yamltext)
        return cls(config, cache, platforms, arches)

    def copy(self):
        """
        Creates a new Installer object with the same configuration
        as this Installer. Necessary to call a nested install().
        """

        return Installer(
            self.descriptor, self.cache, self.platforms, self.arches
        )

    def set_cache_only(self, cache_only):
        """
        If set to true, then calling install() will only execute any
        'url' directives in the corresponding package
        """

        self.cache_only = cache_only

    def set_recache(self, recache):
        """
        If set to true, then calling install() will ignore any previously-
        cached installer files and re-download them (and add the new files
        to the cache, replacing old ones)
        """

        self.recache = recache

    def set_from_local_file(self, from_local_file):
        """
        If a filename is specified here, "cbdep install" will cache and use
        that file rather than downloading anything from the internet. Disables
        recaching.
        """

        self.from_local_file = pathlib.Path(from_local_file)
        self.set_recache(False)
        if not self.from_local_file.exists():
            logger.error(
                f"Specified local file {from_local_file} does not exist!")
            sys.exit(1)

    def get_installer_file(self):
        """
        Returns the (most recent) installer_file, ie, the resulting
        local filename of the most recent 'url' directive
        """

        return self.installer_file

    def install(self, package, version, base_url, inst_dir, force_cbdeps=False):
        """
        Entry point to install a version of named package
        """

        self.package = package
        self.symbols['PACKAGE'] = package
        self.version = version
        self.symbols['VERSION'] = version
        self.base_url = base_url

        # Make install inst_dir absolute
        # QQQ Should use .resolve() rather than .absolute() since the latter
        # is semi-documented and semi-deprecated. However .resolve() doesn't
        # actually work as documented on Windows (doesn't return an absolute
        # path), where .absolute() does. So...
        self.installdir = str(pathlib.Path(inst_dir).absolute())
        self.symbols['INSTALL_DIR'] = self.installdir

        # Determine descriptor block to use for package
        pkgs = self.descriptor.get("packages")
        if pkgs is None:
            logger.error("Malformed configuration file (missing 'packages')")
            sys.exit(1)
        blocks = pkgs.get(package)

        is_cbdeps = False
        if blocks is None:
            # Possibly a cbdeps package, try there
            cbdeps = self.descriptor.get("cbdeps")
            if cbdeps is not None:
                if package in cbdeps['packages'] or force_cbdeps:
                    blocks = cbdeps['descriptor']
                    is_cbdeps = True

        if blocks is None:
            logger.error(f"Unknown package: {package}")
            sys.exit(1)

        # Version number handling. We want the following variables available for
        # templates:
        #
        #  VERSION - exactly what was typed at the command line
        #
        #  VERSION_MAJOR, VERSION_MINOR, VERSION_PATCH, VERSION_BUILD - up to
        #    four numeric components of VERSION
        #
        #  VERSION_MAJORMINORPATCH - the first three numeric components of
        #    VERSION joined with '.' characters
        #
        # We achieve this by splitting VERSION on any non-numeric characters,
        # and then populating an array "version_bits" with up to the first four
        # numbers. In general we do so strictly in order, padding with empty
        # strings; ie, the version is just "1.2", then "version_bits" will be
        # ['1', '2', '', ''].
        #
        # If there are more than four numeric components, for historic reasons
        # we consolidate the excess components into VERSION_PATCH, so that
        # VERSION_BUILD is always a single number.
        #
        # There are two exceptions to those overall rules: OpenJDK and classic
        # cbdeps, which will be handled in-line.
        is_openjdk = package.startswith("openjdk")
        if is_openjdk:
            # OpenJDK has two main exceptions:

            # 1. The build number is always separated with a '+'
            #    character or, in JDK 8, by "-b". Occasionally they also
            #    have a rebuild number, which is represented as an extra
            #    .X after the build number. This rebuild number is only
            #    for occasional repackaging and is not part of the
            #    "real" version, so we do not put it into version_bits.
            #    It is only available via VERSION, which remains exactly
            #    what was typed at the command line - fortunately that
            #    is the only place we need it.
            base_ver, build = re.split(r'\+|-b', self.version)
            build_bits = build.split('.')

            # 2. JDK 8 used "u" to separate MAJOR_VERSION and
            #    MINOR_VERSION, so split on non-numerics rather than
            #    non-alphanumerics.
            version_bits = re.split(r'[^0-9]+', f"{base_ver}+{build_bits[0]}")

        elif is_cbdeps:
            # A few cbdeps packages have version numbers with a "profile"
            # extension, which is separated by a _ character. So for those, we
            # don't want to split on that.
            version_bits = re.split(r'[^A-Za-z0-9_]+', self.version)

        elif package == "java":
            # "java" is a slight exception. Its versioning scheme is
            # truly bizarre; however, cbdep.config actually only cares
            # about VERSION, VERSION_MAJOR, and (only for JDK 8)
            # VERSION_MINOR which needs to be the part after "u" in the
            # JDK 8 versioning system. As above for openjdk, we split on
            # only non-numerics rather than non-alphanumerics. We don't
            # need to do anything different later, so we don't have an
            # is_java boolean here.
            version_bits = re.split(r'[^0-9]+', self.version)

        else:
            # Otherwise, split on any non-alphanumeric characters.
            version_bits = re.split(r'[^A-Za-z0-9]+', self.version)

        # When more than 4 version components are found, we need to do some
        # manipulation to ensure BUILD does not contain the additional info.
        # Instead, we dot-join anything between MINOR and BUILD and treat
        # the combined string as the PATCH component
        if len(version_bits) > 4:
            version_bits[2] = ".".join(version_bits[2:-1])
            version_bits[3] = version_bits[-1]
            version_bits = version_bits[:4]

        # Save a pkg_resources-compatible variant of the version number. Note
        # this will not contain the OpenJDK "rebuild" number if one exists.
        self.safe_version = '.'.join(version_bits)
        logger.debug(f"Safe version is {self.safe_version}")

        # OpenJDK versions were explained earlier. Cbdeps version numbers always
        # have a build number after a hyphen, eg. 71.1-1 or 54.1-cb10. For both
        # OpenJDK and Cbdeps, we want to make sure the final version component
        # is in the "build" slot.
        #
        # For non-OpenJDK/Cbdeps versions, just pad version_bits out to exactly
        # 4 slots.
        num_bits = len(version_bits)
        offset = num_bits - 1 if (is_openjdk or is_cbdeps) else num_bits
        version_bits[offset:0] = [''] * (4 - num_bits)

        self.symbols['VERSION_MAJOR'] = version_bits[0]
        self.symbols['VERSION_MINOR'] = version_bits[1]
        self.symbols['VERSION_PATCH'] = version_bits[2]
        self.symbols['VERSION_BUILD'] = version_bits[3]

        # Also provide single field of major.minor.patch with the appropriate
        # number of dots
        self.symbols['VERSION_MAJORMINORPATCH'] = '.'.join(
            [x for x in version_bits[:3] if x]
        )
        logger.debug(f"Starting install for package {package}")

        block = self.find_block(blocks)
        if block is None:
            logger.error(f"No blocks for package {package} {version} "
                         f"are appropriate for current system")
            sys.exit(1)

        self.execute_block(block)

    def find_block(self, blocks):
        """
        Searches for first block (dictionary) in the list "blocks" which
        is appropriate for the current system, based on its if_platform,
        if_version, etc. keys
        Returns: Said block, or None if none match
        """

        for block in blocks:
            if self.match_platform(block) and \
               self.match_arch(block) and \
               self.match_version(block):
                return block

        return None

    def _match_system(self, block, if_directive, system_values, symbol):
        """
        Common implementation for if_platform and if_arch.
        """
        if if_directive not in block:
            return True

        if_directive_values = block[if_directive]
        if not isinstance(if_directive_values, list):
            if_directive_values = [if_directive_values]

        # Iterate through each directive value and see if the lowercase
        # version of it exists in the list of system values. This logic
        # ensure that cbdep.config authors may control the order that
        # values are checked. In most cases this doesn't matter, but in
        # some corner cases it may. In particular, on Alpine Linux, the
        # system arch values will include eg. both `x64` and
        # `x64-alpine`, where on non-Alpine Linux it would only include
        # `x64`. cbdep.config authors should place the more-specific
        # `x64-alpine` first in the `if_arch` directive to ensure the
        # Alpine variant is matched when it exists.
        matched_value = None
        for directive_value in if_directive_values:
            if directive_value.casefold() in system_values:
                matched_value = directive_value
                break

        if matched_value is not None:
            self.symbols[symbol] = matched_value
            logger.debug(f"Matched {symbol} {matched_value}")

            # Default value for PLATFORM_EXT - kind of a hack to put this here
            # QQQ Allow overriding in config
            if symbol == "PLATFORM":
                if matched_value.startswith(("win", "pc-win")):
                    self.symbols['PLATFORM_EXT'] = "zip"
                    self.symbols['PLATFORM_EXE_EXT'] = ".exe"
                else:
                    self.symbols['PLATFORM_EXT'] = "tar.gz"
                    self.symbols['PLATFORM_EXE_EXT'] = ""
            return True

        return False

    def match_platform(self, block):
        """
        If the block contains an if_platform key, return true if the current
        platform is one of the values for that key, else false.
        If the block does not contain an if_platform key, return true
        """

        return self._match_system(
            block,
            "if_platform",
            self.platforms,
            "PLATFORM"
        )

    def match_arch(self, block):
        """
        If the block contains an if_arch key, return true if the current
        architecture is one of the values for that key, else false.
        If the block contains a default_arches or default_cbdeps_arches key,
        behave as though if_arch existed with a platform-dependent default
        set of arches.
        If the block does not contain either key, return true.
        """

        arches = None
        if "default_arches" in block:
            arches = get_default_arches()
        elif "default_cbdeps_arches" in block:
            arches = get_default_arches(cbdeps_arches=True)
        if arches is not None:
            block["if_arch"] = arches
            logger.debug(f"Set if_arch: {block['if_arch']}")

        return self._match_system(
            block,
            "if_arch",
            self.arches,
            "ARCH"
        )

    def match_version(self, block):
        """
        If the block contains an if_version key, return true if the current
        version matches the value expression, else false.
        If the block does not contain an if_version key, return true
        """
        if "if_version" not in block:
            return True

        # Read the if_version directive
        if_version = block.get("if_version", None)
        if not if_version:
            return True

        return SpecifierSet(if_version).__contains__(self.safe_version)

    def execute_block(self, block):
        """
        Given a single block from the config, execute all actions
        sequentially
        """

        # Set BASE_URL in the symbol table
        if "base_url" in block:
            self.handle_base_url(block.get("base_url"))

        # Enable any environment overrides
        if "set_env" in block:
            self.handle_set_env(block.get("set_env"))

        actions = block.get("actions")
        if actions is None:
            logger.error("Malformed configuration file (missing 'actions')")
            sys.exit(1)

        for action in actions:

            # Special option "fixed_dir" may cause action to be skipped
            if "fixed_dir" in action:
                if self.handle_fixed_dir(action):
                    continue

            if "url" in action:
                self.do_url(action)
            elif self.cache_only:
                # Skip any other actions if doing cache-only
                continue
            elif "cbdep" in action:
                self.do_cbdep(action)
            elif "install_dir" in action:
                self.do_install_dir(action)
            elif "unarchive" in action:
                self.do_unarchive(action)
            elif "raw_binary" in action:
                self.do_raw_binary(action)
            elif "run" in action:
                self.do_run(action)
            else:
                logger.error(
                    "Malformed configuration file (missing action directive)"
                )
                sys.exit(1)

    def handle_set_env(self, env_args):
        """
        Sets values in the cbdep process's environment
        """

        for env_arg in env_args:
            value = self.templatize(env_args[env_arg])
            logger.debug(f"Setting env {env_arg} to {value}")
            os.environ[env_arg] = value

    def handle_base_url(self, url):
        """
        Handles a 'base_url' directive
        """

        # If the user specified a base URL, use that; otherwise use the
        # default base URL specified by the config file
        if self.base_url is not None:
            base_url = self.base_url
            logger.debug(f"Using user-provided base URL {base_url}")
        else:
            base_url = url
            logger.debug(f"Using default base URL {base_url}")
        template = string.Template(base_url)
        self.symbols['BASE_URL'] = template.substitute(**self.symbols)

    def handle_fixed_dir(self, action):
        """
        Handler for 'fixed_dir' option. If this references a directory that
        already exists, presume this action has been completed previously.
        """

        fixed_dir = self.templatize(action["fixed_dir"])
        self.symbols["FIXED_DIR"] = fixed_dir

        return pathlib.Path(fixed_dir).exists()

    def scrape_html(self, localfile, regexp):
        """
        Reads localfile looking for a particular regexp, which is presumed
        to be a valid URL; cache that URL and return the newly-downloaded file
        """

        matcher = re.compile(regexp)
        with open(localfile, encoding='utf-8') as f:
            logger.debug(f"Searching {localfile} for {regexp}...")
            for line in f:
                match = matcher.search(line)
                if match:
                    url = match.group(1)
                    logger.debug(f"...found {url}")
                    return str(self.cache.get(url, self.recache))

        logger.error(f"Scraped HTML did not find {regexp}")
        sys.exit(1)

    def do_url(self, action):
        """
        Handles a 'url' directive
        """

        # Iterate through available URLs; use first successful download.
        # TODO This probably should somehow cache the result using the
        # first URL as a key, or similar. As it is now, if the second URL
        # is the "right" one, then any subsequent runs of cbdep will hit
        # the network first before discovering the right package in the cache.
        urls = action["url"]
        if not isinstance(urls, list):
            urls = [urls]

        exception = None
        for url in urls:
            template = string.Template(url)
            real_url = template.substitute(**self.symbols)

            # If we've been asked to use a local file, here is where we
            # pre-populate the cache
            if self.from_local_file is not None:
                self.cache.put(real_url, self.from_local_file)

            try:
                localfile = self.cache.get(real_url, self.recache)
                break
            except Exception as e:
                exception = e
        else:
            raise exception

        # Handle strange redirects
        if "scrape_html" in action:
            try:
                localfile = self.scrape_html(localfile, action["scrape_html"])
            except:
                # Try again just in case we've cached a bad HTML file
                logger.debug(
                    "Error parsing HTML, trying to get a fresh copy..")
                localfile = self.cache.get(real_url, True)
                localfile = self.scrape_html(localfile, action["scrape_html"])

        # Remember the downloaded file
        self.installer_file = localfile
        self.symbols['DL'] = localfile

    def do_install_dir(self, action):
        """
        Handles an 'install_dir' directive, which resets self.installdir
        """

        template = string.Template(action["install_dir"])
        self.installdir = template.substitute(**self.symbols)
        self.symbols['INSTALL_DIR'] = self.installdir
        logger.info(f"Overriding install dir to {self.installdir}")

    def do_unarchive(self, action):
        """
        Unarchives the downloaded file
        """

        install_dir = pathlib.Path(self.installdir)
        install_dir.mkdir(exist_ok=True, parents=True)

        args = action["unarchive"]

        # The standardized final location for the package. This
        # may already exist!
        if args and "target_dir" in args:
            target_dir_name = self.templatize(args["target_dir"])
        else:
            target_dir_name = f"{self.package}-{self.version}"
        target_dir = install_dir / target_dir_name

        # We extract the archive to a temporary directory
        temp_dir_handle = tempfile.TemporaryDirectory(dir=install_dir)
        temp_dir = pathlib.Path(temp_dir_handle.name)
        unpack_dir = temp_dir / 'unpack'
        unpack_dir.mkdir()
        logger.info(f"Unpacking archive to {target_dir}")

        try:
            shutil.unpack_archive(self.installer_file, unpack_dir)
        except UnicodeEncodeError as e:
            print("ERROR: Extraction failed - please check LANG/LC_ALL in your environment are pointing at character sets inclusive of UTF-8")
            sys.exit(1)

        # Now we want to find the single directory containing the
        # contents we care about from the unpacked archive.
        if args and "toplevel_dir" in args:
            toplevel_dir = self.templatize(args["toplevel_dir"])
            contents_dir = unpack_dir / toplevel_dir
            if not contents_dir.is_dir():
                logger.error(f"Archive does not contain directory {toplevel_dir}!")
                sys.exit(2)
        else:
            # The unpacked directory *itself* is the contents dir
            contents_dir = unpack_dir

        # Sometimes the contents dir needs to be wrapped in a new
        # directory, often "bin"
        if args and "create_toplevel_dir" in args:
            wrap_dir = temp_dir / 'wrap_dir'
            wrap_dir.mkdir()
            contents_dir.rename(
                wrap_dir / self.templatize(args["create_toplevel_dir"]))
            contents_dir = wrap_dir

        # Finally as atomically as possible, move the existing
        # target directory out of the way (if it exists) and move
        # the contents directory to the target directory.
        if target_dir.exists():
            target_dir.rename(temp_dir / "recycle")
        contents_dir.rename(target_dir)

    def do_raw_binary(self, action):
        """
        Handles a `raw_binary` directive - a single binary download,
        which we will place into a `bin` directory. By default, the
        executable will be given the same name as the file created by
        the download. If a different name is desired or the download URL
        doesn't have a reasonable name, the target executable name can
        be specified in the `name` field.
        """

        install_dir = pathlib.Path(self.installdir)
        install_dir.mkdir(exist_ok=True, parents=True)

        args = action["raw_binary"]
        if args and "name" in args:
            template = string.Template(args["name"])
            bin_name = template.substitute(**self.symbols)
        else:
            bin_name = self.installer_file.name

        # The standardized final location for the package. This may
        # already exist! Note that we don't support a target_dir option
        # here as it should be unnecessary.
        target_dir_name = f"{self.package}-{self.version}"
        target_dir = install_dir / target_dir_name
        logger.info(f"Copying binary to {target_dir}/bin")

        # Create a temporary directory
        temp_dir_handle = tempfile.TemporaryDirectory(dir=install_dir)
        temp_dir = pathlib.Path(temp_dir_handle.name)

        # Create a `bin` directory in the temp dir and copy the binary,
        # ensuring it's executable
        contents_dir = temp_dir / 'contents'
        bin_dir = contents_dir / 'bin'
        logger.debug(f"Creating {bin_dir}")
        bin_dir.mkdir(parents=True)
        bin_file = bin_dir / bin_name
        logger.debug(f"Copying {self.installer_file} to {bin_file}")
        shutil.copy(self.installer_file, bin_file)
        bin_file.chmod(bin_file.stat().st_mode | stat.S_IEXEC)

        # As atomically as possible, move the existing target directory
        # out of the way (if it exists) and move the new directory into
        # place.
        if target_dir.exists():
            target_dir.rename(temp_dir / "recycle")
        contents_dir.rename(target_dir)

    def do_cbdep(self, action):
        """
        Runs a nested "cbdep install" command
        """

        package = action["cbdep"]
        version = action["version"]
        install_dir = self.templatize(
            action.get("install_dir", self.installdir))

        installer = self.copy()
        logger.info(
            f"Calling nested cbdep install -d {install_dir} {package} {version}")
        installer.install(package, str(version), self.base_url, install_dir)

    def do_run(self, action):
        """
        Runs a sequence of local commands from a 'run' digrective
        """

        command_string = self.templatize(action["run"])

        for command in command_string.splitlines():
            logger.debug(f"Running local command: {command}")
            try:
                run(command, shell=True, check=True)
            except CalledProcessError as e:
                # Command failed - error message should already be on stderr
                # Exit cleanly without showing Python traceback
                logger.error(f"Command failed with exit code {e.returncode}")
                sys.exit(e.returncode)

    def templatize(self, template):
        """
        Utility function for doing template substitution
        """

        return string.Template(template).substitute(**self.symbols)



import os
import sys
import zipfile
from shutil import ReadError, _ensure_directory, copyfileobj, register_unpack_format, unregister_unpack_format

"""
Works around two failings in the Python standard library:
1. shutil.extract_archive() does custom handling for zips rather than
   just calling zipfile.extractall()
2. zipfile itself makes note of file permissions in the .zip but does
   not apply them on unpacking
"""

class ZipFileWithPermissions(zipfile.ZipFile):
    """
    Custom ZipFile class handling file permissions.
    """
    def _extract_member(self, member, targetpath, pwd):
        if not isinstance(member, zipfile.ZipInfo):
            member = self.getinfo(member)

        targetpath = super()._extract_member(member, targetpath, pwd)

        # Only attempt permission fixes if the zip file reports itself
        # as being created on a Unix-like system
        if member.create_system == 3:
            attr = member.external_attr >> 16
            if attr != 0:
                os.chmod(targetpath, attr)

        return targetpath

def _unpack_zipfile_with_permissions(filename, extract_dir):
    """
    Unpack zip `filename` to `extract_dir`
    """
    if not zipfile.is_zipfile(filename):
        raise ReadError("%s is not a zip file" % filename)

    with ZipFileWithPermissions(filename) as zip:
        zip.extractall(path=extract_dir)

def register():
    """
    Configures shutil.unpack_archive to use our custom unzipper
    """
    if sys.platform != "win32":
        unregister_unpack_format('zip')
        register_unpack_format('zip', ['.zip'], _unpack_zipfile_with_permissions)



import pytest
import requests
import sys
import tempfile
from hashlib import md5
from pathlib import Path
from shutil import rmtree
from urllib.parse import urljoin
sys.path.append('../scripts')
from cbdep.cache import Cache

dep_url_path = "https://packages.couchbase.com/couchbase-server/deps/openssl/1.1.1l/1/"

filename = {
    "centos": "openssl-centos7-x86_64-1.1.1l-1.tgz",
    "ubuntu": "openssl-ubuntu20.04-x86_64-1.1.1l-1.tgz"
}
url = {
    "centos": urljoin(dep_url_path, filename["centos"]),
    "ubuntu": urljoin(dep_url_path, filename["ubuntu"])
}
hash = {
    "centos": "d8ba3ac17a080a9cde889fdb26308bb8",
    "ubuntu": "59bfb1a1f49ab9e04e9993b3f0bcdc23"
}
name_sha = {
    "centos": md5(url["centos"].encode('utf-8')).hexdigest(),
    "ubuntu": md5(url["ubuntu"].encode('utf-8')).hexdigest()
}

cachetempdir = tempfile.TemporaryDirectory()
cachedir = Path(cachetempdir.name)

class TestCache:
    cache = Cache(cachedir)

    def test__cachefilename(self):
         assert self.cache._cachefilename(cachedir) == cachedir/"filename"

    def test__cachedir(self):
        rmtree(cachedir/name_sha["centos"][0:2], ignore_errors=True)
        path = cachedir/name_sha["centos"][0:2]/name_sha["centos"]
        urlfile = path / "url"
        assert self.cache._cachedir(url["centos"]) == path
        assert open(urlfile).read() == url["centos"]

    def test__writefilename(self):
        self.cache._writefilename(cachedir, "foo")
        assert open(cachedir/"filename").read() == "foo"
        self.cache._writefilename(cachedir, "bar")
        assert open(cachedir/"filename").read() == "bar"

    def test_save(self):
        self.cache.save(url["centos"], cachedir/"dummy")
        assert md5(open(cachedir/"dummy", "rb").read()).hexdigest() == hash["centos"]

    def test_report(self):
        from contextlib import redirect_stdout
        from io import StringIO
        stdout = StringIO()
        with redirect_stdout(stdout):
            self.cache.report(url["centos"])
        assert stdout.getvalue().strip() == str(cachedir/name_sha["centos"][0:2]/name_sha["centos"]/url["centos"].split("/")[-1])

    def test_put(self):
        rmtree(cachedir, ignore_errors=True)
        cachedir.mkdir()
        r = requests.get(url["ubuntu"])
        with open(cachedir/"openssl-ubuntu20.04-x86_64-1.1.1l-1.tgz", 'wb') as f:
            f.write(r.content)
        self.cache.put(url["centos"], cachedir/"openssl-ubuntu20.04-x86_64-1.1.1l-1.tgz")
        assert open(cachedir/name_sha["centos"][0:2]/name_sha["centos"]/"filename").read() == filename["ubuntu"]
        assert open(cachedir/name_sha["centos"][0:2]/name_sha["centos"]/"url").read() == url["centos"]

    def test_get(self):
        rmtree(cachedir, ignore_errors=True)
        cachedir.mkdir()
        self.cache.get(url["ubuntu"])
        assert md5(open(cachedir/name_sha["ubuntu"][0:2]/name_sha["ubuntu"]/filename["ubuntu"], "rb").read()).hexdigest() == hash["ubuntu"]
        self.cache.get(url["ubuntu"])
        assert md5(open(cachedir/name_sha["ubuntu"][0:2]/name_sha["ubuntu"]/filename["ubuntu"], "rb").read()).hexdigest() == hash["ubuntu"]
        self.cache.get(url["ubuntu"], recache=True)
        assert md5(open(cachedir/name_sha["ubuntu"][0:2]/name_sha["ubuntu"]/filename["ubuntu"], "rb").read()).hexdigest() == hash["ubuntu"]



import pytest
import subprocess
import sys
import tempfile

cbdep_install_tests = [
    ["uv", "0.4.29", "unknown-linux-gnu"],
    ["openjdk", "9.0.4+11"],
    ["openjdk", "8u282-b08"],
    ["openjdk", "11.0.9.1+1"],
    ["openjdk", "11.0.9+11"],
    # openjdk with a dotted build number (CBD-5669)
    ["openjdk", "17.0.9+9.1", "windows"],
    ["openjdk-jre", "9.0.4+11"],
    ["openjdk-jre", "8u282-b08"],
    ["openjdk-jre", "11.0.9.1+1"],
    ["openjdk-jre", "11.0.9+11"],
    ["corretto", "11.0.21.9.1"],
    ["python", "3.9.1"],
    ["miniforge3", "22.9.0-1"],
    ["mambaforge", "4.13.0-1"],
    ["miniconda3-py38", "4.12.0"],
    ["miniconda3-py39", "4.12.0"],
    ["miniconda2", "4.5.11"],
    ["cmake", "3.25.0-rc4"],
    ["cmake", "3.24.3"],
    ["ninja", "1.11.1"],
    # check_one_package wix 3.11.2 windows # windows only
    ["jq", "1.7"],
    ["helm", "3.17.0"],
    ["flux", "2.5.1"],
    ["prometheus", "2.40.1"],
    ["golang", "1.19.3"],
    ["dotnet-core-sdk", "7.0.100"],
    ["dotnet-core-runtime", "6.0.11"],
    ["docker", "20.10.21"],
    ["nodejs", "18.12.1"],
    # check_one_package php 7.3.4-cb1 - legacy package, can't find any installation candidates?
    ["php-nts", "8.1.4-cb1"],
    ["php-zts", "8.1.4-cb1"],
    ["libcouchbase_vc11", "2.9.5", "windows"],
    ["libcouchbase_vc14", "3.0.0", "windows"],
    ["couchbasemock", "1.5.25"],
    ["analytics-jars", "7.0.5-7643"],
]

install_dir = tempfile.TemporaryDirectory()

class TestCbdepInstall:

    @pytest.fixture(params=cbdep_install_tests)
    def args(self, request):
        return request.param

    def test_cbdep_install(self, args):

        if len(args) > 2:
            platform = args[2]
        else:
            platform = "macosx" if sys.platform == "darwin" else "linux"

        subprocess.run(
            [
                "uv", "run", "cbdep",
                "--platform", platform,
                "install", "-d", install_dir.name, args[0], args[1]
            ],
            check=True
        )



"""
Functions for determining current platform information
"""

import platform

_platforms = None
_arch = None

_intel_arches = ["x86_64", "amd64", "x64"]
_arm_arches = ["aarch64", "arm64", "aarch_64"]

def override_platforms(platforms):
    """
    Set platform to a specific value or values, to suppress introspection
    """

    global _platforms
    _platforms = platforms

def override_arch(arch):
    """
    Set initial architecture to specific value - additional synonyms may
    still be added
    """

    global _arch
    _arch = arch

def get_platforms():
    """
    Returns a list of increasingly-generic identifiers for the current
    system.
    """

    global _platforms
    if _platforms is not None:
        return _platforms

    # Start with the most generic - the OS
    system = platform.system().lower()

    # Initialize list with the system.
    _platforms = [system]

    # OS-specific stuff
    if system == "linux":
        import distro

        # Add distro-specific variants
        dist_id = distro.id()
        _platforms.insert(0, dist_id)

        if dist_id == "ubuntu":
            # Ubuntu "minor" versions are distinct, eg., Ubuntu 16.10
            # is potentially quite different from Ubuntu 16.04. So we
            # want to use the combined version.
            dist_ver = distro.version()
        else:
            # Other supported distros use rolling releases, so eg.
            # Centos 6.5 shouldn't differ importantly from Centos 6.9.
            # Use only the major version number.
            dist_ver = distro.major_version()

        _platforms.insert(0, f"{dist_id}{dist_ver}")
        _platforms.insert(0, f"{dist_id}-{dist_ver}")

        if dist_id == "sles" or dist_id.startswith("opensuse"):
            # Cbdeps 1.0, at least, refers to all SUSE as "suse", so offer
            # those as platform names too
            dist_id = "suse"
            _platforms.insert(0, dist_id)
            _platforms.insert(0, f"{dist_id}{dist_ver}")
            _platforms.insert(0, f"{dist_id}-{dist_ver}")

        # Add the "target triple" version
        if dist_id == "alpine":
            _platforms.insert(0, "unknown-linux-musl")
        else:
            _platforms.insert(0, "unknown-linux-gnu")

    elif system == "darwin":
        _platforms.insert(0, "apple-darwin")
        _platforms.insert(0, "macosx")
        _platforms.insert(0, "macos")
        _platforms.insert(0, "mac")
        _platforms.insert(0, "osx")

    elif system == "windows":
        # QQQ Somehow introspect MSVC version?
        _platforms.insert(0, "windows_msvc2015")
        _platforms.insert(0, "windows_msvc2017")
        _platforms.insert(0, "pc-windows-msvc")
        _platforms.insert(0, "win")

    return _platforms

def _alpine_mod(arches):
    """
    If this is Alpine Linux, append "musl" and "alpine" to each arch
    entry in arches. This appears to be a common convention (at least
    nodejs, dotnet, and Bellsoft OpenJDK use it).
    """

    if "alpine" not in get_platforms():
        return arches.copy()
    # Ensure the -musl and -alpine variants appear first for all arches
    # - we don't want, eg., a plain "x86_64" to be on the list before
    # "x64-alpine"
    return [f"{arch}{suffix}"
        for suffix in ["-musl", "-alpine", ""]
        for arch in arches
    ]

def get_arches():
    """
    Returns a list of possible architectures based on the current system
    """

    global _arch, _intel_arches, _arm_arches

    # Start from "machine" type (possibly overridden)
    if _arch is not None:
        arch = _arch
    else:
        arch = platform.machine().casefold()

    # If we recognize this as being Intel or ARM, return a set of common
    # synonyms
    if arch in _intel_arches:
        return _alpine_mod(_intel_arches)
    if arch in _arm_arches:
        return _alpine_mod(_arm_arches)
    # Otherwise, just return what we've got
    return _alpine_mod([arch])

def get_default_arches(cbdeps_arches=False):
    """
    Returns a platform-dependent value suitable for if_arch directives.
    Important: this is a list of arch names most frequently used in *package*
    filenames for a given OS. It should *not* contain any "synonyms", like
    "arm64" and "aarch64", or "x86_64" and "amd64", because given the way
    if_arch works, only the first will ever get matched. Basically these
    should contain just the most common name (if any) for "32-bit Intel",
    "64-bit Intel", "64-bit ARM", and any other supported arches on each OS.
    """

    for plat in get_platforms():
        if plat.startswith("win"):
            if cbdeps_arches:
                return ["x86", "amd64", "arm64"]
            else:
                return ["x86", "x86_64", "arm64"]
        elif "darwin" in plat or "mac" in plat:
            return ["x86_64", "arm64"]
        elif "android" in plat:
            return ["armv7a", "aarch64", "i686", "x86_64"]

    # Didn't find anything else, so return the list for Linux, including Alpine
    return ["x86", "x86_64", "aarch64", "x64-musl"]



"""
Cache
"""

import hashlib
import logging
import os
import pathlib
import re
import requests
import shutil
import urllib.parse

# Set up logging and handler
logger = logging.getLogger('cbdep')


class Cache:
    """
    Class to store downloaded files in a local cache, keyed by URL.

    Local cache directory structure is keyed on an MD5 checksum of the
    URL, and contains directories named:
        <first 2 chars of checksum> / <checksum> /
    In each directory will be three files:
        url - contains the input URL
        filename - contains the filename associated with the download
        [filename] - contains the downloaded contents
    """

    def __init__(self, directory):
        """
        Initialize a cache based at the specified directory
        """
        self.directory = pathlib.Path(directory)

    def get(self, url, recache=False):
        """
        Downloads url (if necessary), saves in local cache. If recache is
        True, will always re-download the url.
        """

        cachedir = self._cachedir(url)
        cachefilename = self._cachefilename(cachedir)

        # If "filename" file exists, it's a hit; read the actual filename
        # from there and return the cached content file
        if cachefilename.exists() and not recache:
            logger.debug(f"Cache hit for {url}")
            with open(cachefilename) as f:
                filename = f.readline()
                return cachedir / filename

        # Cache miss; attempt to download the URL
        with requests.get(url, allow_redirects=True, stream=True,
                          timeout=30.0) as r:
            r.raise_for_status()

            # Determine download filename
            filename = None
            cd = r.headers.get('content-disposition')
            if cd:
                filenames = re.findall('filename=([^;]+)', cd)
                if len(filenames) > 0:
                    filename = filenames[0]
            if filename is None:
                filename = os.path.basename(urllib.parse.urlparse(url).path)
            logger.info(f"Caching {url} ({filename})")

            cachefile = cachedir / filename
            try:
                # Download file
                with open(cachefile, 'wb') as fd:
                    for chunk in r.iter_content(chunk_size=1024):
                        fd.write(chunk)

                self._writefilename(cachedir, filename)

            except:
                if cachefile.exists():
                    cachefile.unlink()
                if cachefilename.exists():
                    cachefilename.unlink()
                raise

        logger.debug("Downloaded file")
        return cachefile

    def put(self, url, localfile):
        """
        Copies the specified local pathlib handle into the cache keyed by the
        specified URL.
        Returns pathlib handle to cached file.
        """

        cachedir = self._cachedir(url)
        filename = localfile.name

        logger.debug(f"Storing {localfile} in cache for {url}")
        shutil.copy2(localfile, cachedir / filename)
        self._writefilename(cachedir, filename)

    def report(self, url):
        """
        Output the full path to the file in the cache
        """

        print(self.get(url))

    def save(self, url, output):
        """
        Save a local copy of a file in the cache
        """

        shutil.copy2(self.get(url), output)

    def _cachedir(self, url):
        """
        Returns pathlib handle to cache directory for given URL. Creates cachedir
        if necessary, with initial "url" file entry.
        """

        md5 = hashlib.md5(url.encode('utf-8')).hexdigest()
        cachedir = self.directory / md5[0:2] / md5

        if not cachedir.exists():
            logger.debug(f"Creating cache directory {cachedir}")
            cachedir.mkdir(parents=True)
            with open(cachedir / "url", 'w') as f:
                f.write(url)

        return cachedir

    def _writefilename(self, cachedir, filename):
        """
        Writes the "filename" file into the appropriate cache directory. As a
        side effect, if "filename" already exists and references a different
        file, that (now stale) file will be removed.
        """

        cachefilename = self._cachefilename(cachedir)
        if cachefilename.exists():
            with open(cachefilename) as f:
                currentfilename = f.readline()
            if currentfilename != filename:
                cachedfile = cachedir / currentfilename
                if cachedfile.exists():
                    cachedfile.unlink()

        logger.debug(f"Recording filename {filename}")
        with open(cachefilename, 'w') as f:
            f.write(filename)

    def _cachefilename(self, cachedir):
        """
        Returns pathlib handle to the "filename" file in cachedir
        """

        return cachedir / "filename"



import importlib
import logging
import os
import sys
import pytest
import tempfile
from hashlib import md5
from pathlib import Path
from shutil import rmtree
sys.path.append('../scripts')
from cbdep.cache import Cache
from cbdep.install import Installer
import cbdep.platform_introspection as plat
import cbdep

wrong_platform_package = {
    "name": "java",
    "platform": "stegosaurus",
    "version": "11.0.3",
    "filename": "jdk-11.0.3_linux-x64_bin.tar.gz",
    "hash": "9bdefa0ac6378798643934ea6a065fe1"
}

missing_package = {
    "name": "foo",
    "platform": "linux",
    "version": "0.0.0",
    "hash": "0000",
}

def package_name(package):
    return package.get("name")

# The `platform` and `arch` values are used as the cbdep-provided values
# for the environment cbdep is being run in.
install_packages = [
    {
        "name": "dotnet-core-runtime",
        "platform": "linux",
        "version": "5.0.10",
        "arch": "x64",
        "filename": "dotnet-runtime-5.0.10-linux-x64.tar.gz",
        "hash": "7bb4e01127c7188bb6f4bfe201f1aa20",
        "final_file": "dotnet"
    },
    {
        # Something for Alpine
        "name": "dotnet-core-runtime",
        "platform": "linux",
        "version": "6.0.0",
        # This arch array is actually backwards compared to the version
        # that would be introspected on Alpine Linux; we use this to
        # test that the `if_arch` values are checked in the order
        # specified by `cbdep.config`.
        "arch": ["x64", "x64-alpine"],
        "filename": "dotnet-runtime-6.0.0-linux-musl-x64.tar.gz",
        "hash": "745119ce18e8e0e73b3d44e000d23621",
        "final_file": "dotnet"
    },
    {
        # Another thing for Alpine
        "name": "nodejs",
        "platform": "linux",
        "version": "16.20.2",
        "arch": ["x64", "x64-musl"],
        "filename": "node-v16.20.2-linux-x64-musl.tar.gz",
        "hash": "90dc2aa03bd10272452e4eae11b82e66",
        "final_file": "bin/node"
    },
    {
        "name": "golang",
        "platform": "linux",
        "arch": "amd64",
        "version": "1.24.7",
        "filename": "go1.24.7.linux-amd64.tar.gz",
        "install_dir": "go1.24.7",
        "hash": "fba8e1e6955bd708025e8db52b0f14ea",
        "final_file": "bin/go"
    },
    {
        # random cbdep
        "name": "analytics-jars",
        "base_url": "https://packages.couchbase.com/releases",
        "platform": "linux",
        "version": "7.0.2-6512",
        "hash": "0bbb2376f42848451b6898247aa85793",
        "final_file": "jars/ini4j-0.5.4.jar"
    },
    {
        "name": "openjdk",
        "platform": "linux",
        "version": "11.0.12+7",
        "filename": "OpenJDK11U-jdk_x64_linux_hotspot_11.0.12_7.tar.gz",
        "hash": "4402d15b299b48172d1577b374392e63",
        "final_file": "bin/java"
    },
    {
        # MB-54306: openjdk with optional version segment
        "name": "openjdk",
        "platform": "linux",
        "version": "11.0.16.1+1",
        "filename": "OpenJDK11U-jdk_x64_linux_hotspot_11.0.16.1_1.tar.gz",
        "hash": "251cbce7b80e8fdcb48285867bba77d6",
        "final_file": "bin/java"
    },
    {
        # CBD-5669: openjdk with rebuild number
        "name": "openjdk",
        "platform": "windows",
        "version": "17.0.9+9.1",
        "filename": "OpenJDK17U-jdk_x64_windows_hotspot_17.0.9_9.zip",
        "hash": "4e45d6412a88b2dbef88d5391d533a46",
        "final_file": "bin/java.exe"
    },
    {
        # classic cbdep - also tests padding of version_bits
        "name": "icu4c",
        "base_url": "https://packages.couchbase.com/couchbase-server/deps",
        "platform": "amzn2",
        "version": "59.1-cb2",
        "filename": "icu4c-amzn2-x86_64-59.1-cb2.tgz",
        "hash": "6e95a3aa1dcd0f38c10596244e500c9d",
        "final_file": "include/unicode/ustring.h"
    },
    {
        # classic cbdep with "profile" extension in version number
        "name": "curl",
        "base_url": "https://packages.couchbase.com/couchbase-server/deps",
        "platform": "linux",
        "version": "8.4.0-1_openssl31x",
        "filename": "curl-linux-x86_64-8.4.0-1_openssl31x.tgz",
        "hash": "7bc91f2d00d89e7fda734556d0121946",
        "final_file": "lib/libcurl.so.4.8.0"
    },
    {
        # something with install_dir
        "name": "php-zts",
        "base_url": "https://packages.couchbase.com/couchbase-server/deps/php",
        "platform": "linux",
        "version": "8.0.2-cb1",
        "install_dir": "/tmp/php/php-zts-8.0.2-cb1",
        "filename": "php-zts-linux-x86_64-8.0.2-cb1.tgz",
        "hash": "fe52a6efd938d5d3ed4c5b9031477fe9",
        "final_file": "bin/php"
    },
    {
        # something with run
        "name": "docker",
        "platform": "linux",
        "version": "20.10.3",
        "filename": "docker-20.10.3.tgz",
        "hash": "4bb3ff2457b540995ddae3124f06d40a",
        "final_file": "bin/docker"
    },
    {
        # something with both toplevel_dir and create_toplevel_dir
        "name": "uv",
        "platform": "unknown-linux-gnu",
        "version": "0.4.29",
        "filename": "uv-x86_64-unknown-linux-gnu.tar.gz",
        "hash": "3c08a4629ca3368f80e26a091a6f1496",
        "final_file": "bin/uvx"
    },
    {
        # something with raw_binary
        "name": "kubectl",
        "platform": "linux",
        "version": "1.31.0",
        "filename": "kubectl",
        "hash": "d2e1587ead4510060de349c9dc8b7861",
        "final_file": "bin/kubectl"
    }
]

yamltext = (importlib.resources.files(cbdep) / "cbdep.config").read_text()
tempdir = tempfile.TemporaryDirectory()
wd = Path(tempdir.name)
logger = logging.getLogger()

def clear_wd():
    rmtree(wd, ignore_errors=True)

class TestInstaller:

    def test_fromYaml(self):
        installer = Installer.fromYaml(yamltext, Cache(wd), "linux", "x86_64")
        assert type(installer) == Installer

    def test___del__(self):
        installer = Installer.fromYaml(yamltext, Cache(wd), "linux", "x86_64")
        tempdir = Path(installer.temp_dir)
        tempdir.mkdir(exist_ok=True)
        assert tempdir.exists()
        del installer
        assert tempdir.exists() == False

    def test_copy(self):
        installer = Installer.fromYaml(yamltext, Cache(wd), "linux", "x86_64")
        assert type(installer.copy()) == Installer

    def test_set_cache_only(self):
        installer = Installer.fromYaml(yamltext, Cache(wd), "linux", "x86_64")
        installer.set_cache_only(False)
        assert installer.cache_only == False
        installer.set_cache_only(True)
        assert installer.cache_only == True

    def test_recache(self):
        installer = Installer.fromYaml(yamltext, Cache(wd), "linux", "x86_64")
        installer.set_recache(False)
        assert installer.recache == False
        installer.set_recache(True)
        assert installer.recache == True

    def test_set_from_local_file(self):
        installer = Installer.fromYaml(yamltext, Cache(wd), "linux", "x86_64")
        Path("/tmp/bar").touch()
        installer.set_recache(True)
        installer.set_from_local_file("/tmp/bar")
        assert installer.recache == False
        with pytest.raises(SystemExit) as e:
                installer.set_from_local_file("/tmp/barbaz")
        assert e.type == SystemExit
        assert e.value.code == 1

    def test_get_installer_file(self):
        installer = Installer.fromYaml(yamltext, Cache(wd), "linux", "x86_64")
        installer.installer_file = "foo"
        assert installer.get_installer_file() == "foo"

    def test_templatize(self):
        installer = Installer.fromYaml(yamltext, Cache(wd), "linux", "x86_64")
        installer.symbols["ALPHA"] = "ABC"
        installer.symbols["NUMERIC"] = "123"
        assert installer.templatize("${ALPHA}-${NUMERIC}") == "ABC-123"

    # Fixture to parameterize test_working_install() for each package config
    @pytest.fixture(params=install_packages, ids=package_name)
    def package(self, request):
        return request.param

    def test_working_install(self, caplog, package):
        caplog.set_level(logging.DEBUG)
        clear_wd()
        logger.info(f"Testing package '{package}'")
        installer = Installer.fromYaml(yamltext, Cache(wd), package["platform"], package.get("arch", plat.get_arches()))
        installer.install(package["name"], package["version"], package.get("base_url", ""), wd/"install")
        fn = wd / package["hash"][0:2] / package["hash"] / package.get("filename", f"{package['name']}-{package['version']}.tar.gz")
        logger.debug(f"    Checking for downloaded file '{fn}'")
        assert fn.is_file()
        install_dir = Path(package.get("install_dir", f"{package['name']}-{package['version']}"))
        if not install_dir.is_absolute():
            install_dir = wd / "install" / install_dir
        logger.debug(f"    Checking for install dir '{install_dir}'")
        assert install_dir.is_dir()
        final_filename = package.get("final_file", None)
        if final_filename is not None:
            logger.debug(f"    Checking for final file '{final_filename}'")
            assert (install_dir / final_filename).is_file()
        rmtree(install_dir, ignore_errors=True)

    def test_broken_install(self):
        clear_wd()
        installer = Installer.fromYaml(yamltext, Cache(wd), missing_package["platform"], "x86_64")
        with pytest.raises(SystemExit) as e:
                installer.install(missing_package["name"], missing_package["version"], missing_package.get("base_url", ""), wd/"install")
        assert e.type == SystemExit
        assert e.value.code == 1

    def test_wrong_platform_install(self):
        clear_wd()
        installer = Installer.fromYaml(yamltext, Cache(wd), wrong_platform_package["platform"], "x86_64")
        with pytest.raises(SystemExit) as e:
            installer.install(wrong_platform_package["name"], wrong_platform_package["version"], wrong_platform_package.get("base_url", ""), wd/"install")
        assert e.type == SystemExit
        assert e.value.code == 1

    def test_handle_set_env(self):
        installer = Installer.fromYaml(yamltext, Cache(wd), "linux", "x86_64")
        installer.symbols["foo"] = "bar"
        os.environ["test_handle_set_env"] = ""
        installer.handle_set_env({ "test_handle_set_env": "xxx" })
        assert os.environ["test_handle_set_env"] == "xxx"

    def test_handle_fixed_dir(self):
        rmtree(wd, ignore_errors=True)
        wd.mkdir()
        installer = Installer.fromYaml(yamltext, Cache(wd), "linux", "x86_64")
        installer.symbols["foo"] = "bar"
        installer.handle_fixed_dir({ "fixed_dir": str(wd/"missing") })
        assert installer.handle_fixed_dir({ "fixed_dir": str(wd/"missing") }) == False
        (wd/"present").touch()
        assert installer.handle_fixed_dir({ "fixed_dir": str(wd/"present") }) == True

    def test_do_install_dir(self):
        installer = Installer.fromYaml(yamltext, Cache(wd), "linux", "x86_64")
        action = {
            "install_dir": str(wd / "test_do_install_dir")
        }
        installer.do_install_dir(action)
        assert installer.symbols["INSTALL_DIR"] == str(wd / "test_do_install_dir")

    def test_do_cbdep(self):
        rmtree(wd, ignore_errors=True)
        wd.mkdir()
        assert yamltext
        installer = Installer.fromYaml(yamltext, Cache(wd), "linux", "x86_64")
        installer.do_cbdep({"cbdep": "analytics-jars", "version": "7.0.2-6512", "install_dir": str(wd/"test_do_cbdep")})
        assert md5(open(wd/"test_do_cbdep"/"analytics-jars-7.0.2-6512"/"cbas-install-7.0.2.jar", "rb").read()).hexdigest() == "3436fda4756c9aed996a6ad2ed9ddb30"



import argparse
import cbdep.cli as cli

class TestList:
    def test__list(self):
        parser = argparse.ArgumentParser()
        parser.add_argument("--config-file", type=str)
        args = parser.parse_args()
        tool = cli.Cbdep()
        tool.do_list(args)
